// Generated from Flocc program
// see http://www.flocc.net/

#ifndef HEADER_H
#define HEADER_H

#include <cmath>
#include <mpi.h>

//#define DEBUG 1
#include "pointer.h"
#include "hashing.h"
#include "vecmaps.h"
#include "reparter.h"
#include "carttops.h"

// BASE TYPES 

typedef struct nullStruct {
  bool operator<(const nullStruct& b) const {
    return false;
  }
  bool operator==(const nullStruct& b) const {
    return true;
  }
} nullStruct;

// PRINTING

void printVal(int v) { printf("%i", v); }
void printVal(const char *v) { printf("%s", v); }
void printVal(nullStruct v) {
  printVal("null");
}
void printVal(int* v) {
  // only safe to print one, as don't know array length
  printVal(*v); 
} 

// MPI STUFF

// converts the rank of a process in one communicator to it's rank
// in another communicator.
void translateRank(MPI::Comm *srcComm, int srcRank, MPI::Comm *destComm, int *destRank) {
  MPI_Group orig_group, new_group; 
  orig_group = srcComm->Get_group();
  new_group = destComm->Get_group();
  MPI::Group::Translate_ranks(orig_group, 1, &srcRank, new_group, destRank);
} 

//typedef void MPI::Comm::Errhandler_fn(MPI::Comm &, int *, ... );
void throwErrHandler(MPI::Comm &comm, int *errCode, ...) {
  int len = 5000;
  char str[len];
  for (int i = 0; i < len; i++) str[i] = 0;
  MPI::Get_error_string(*errCode, str, len);
  std::cerr << "MPI Error: " << str << "\n";
  throw 1;
}

// TIME

#define MAX_TIME_HEADER 1024
int timeRank;
char timeHeader[MAX_TIME_HEADER];
double beginTime, middleTime;

void initTime() {
  // rank
  timeRank = MPI::COMM_WORLD.Get_rank();

  // start time
  beginTime = MPI_Wtime();
  middleTime = beginTime;
}

void logTime(const char* event, double *startTime) {
  if (timeRank == 0) {
    double End = MPI_Wtime();
    std::cerr << timeHeader << "\t" 
              << event << "\t" << (End - (*startTime)) << std::endl;
  }
  middleTime = MPI_Wtime();
}

#endif

// Global cartesian communicator
static const int ndims = 1;
static const int maxNodesPerDim = 4096;
static std::vector<int> dimSizes;
static MPI::Cartcomm cartComm;
static int coordRanks[maxNodesPerDim];
static int thisCoord[ndims];
static int thisRank, Nproc, rootRank;


typedef struct struct10 {
double v0;
double v1;
bool operator<(const struct10& b) const {
  return (v0 < b.v0) || ((v0 == b.v0) && (((v1 < b.v1))));
}
bool operator>=(const struct10& b) const {
  return !((v0 < b.v0) || ((v0 == b.v0) && (((v1 < b.v1)))));
}
inline bool operator==(const struct10& b) const {
  return (v0 == b.v0) && (v1 == b.v1);
}
inline bool operator!=(const struct10& b) const {
  return !((v0 == b.v0) && (v1 == b.v1));
}
} struct10;
void printVal(const struct10& v) {
printVal("(");
printVal(v.v0);
printVal(", ");
printVal(v.v1);
printVal(")");

}
 namespace flocc {
   template<>
   struct hasher<struct10> {
   public:
       static inline hash_t hash(struct10 const& v) {
           hash_t seed = 0;
             hash_combine<double>(seed, v.v0);
  hash_combine<double>(seed, v.v1);
  
           return seed;
        }
   };
 }
struct eqstruct10 {
inline bool operator()(const struct10& a, const struct10& b) const {
  return (a.v0 == b.v0) && (a.v1 == b.v1);
}
};


 // user def reduction function
  void userFun400(const void* invec, void *inoutvec, int len, const MPI::Datatype& datatype) {
   const struct10* in = (const struct10*)invec;
   struct10* out = (struct10*)inoutvec;
   // TODO: WHEN SENDING PROPER DATATYPE, CHANGE (len / sizeof(struct10)) to len!
   for (int i = 0; i < (len / sizeof(struct10)); i++) {
     struct10 val2 = out[i];
         double v416;
    v416 = in[i].v1 + val2.v1;
    double v417;
    v417 = in[i].v0 + val2.v0;
  // TupNd assignment
    out[i].v0 = v417; //(Lf:(Float )) = (Lf:(Float ))
    out[i].v1 = v416; //(Lf:(Float )) = (Lf:(Float ))
    
  
  }
 }

MPI::Op v418;

 // user def reduction function
  void userFun450(const void* invec, void *inoutvec, int len, const MPI::Datatype& datatype) {
   const struct10* in = (const struct10*)invec;
   struct10* out = (struct10*)inoutvec;
   // TODO: WHEN SENDING PROPER DATATYPE, CHANGE (len / sizeof(struct10)) to len!
   for (int i = 0; i < (len / sizeof(struct10)); i++) {
     struct10 val2 = out[i];
         double v466;
    v466 = in[i].v1 + val2.v1;
    double v467;
    v467 = in[i].v0 + val2.v0;
  // TupNd assignment
    out[i].v0 = v467; //(Lf:(Float )) = (Lf:(Float ))
    out[i].v1 = v466; //(Lf:(Float )) = (Lf:(Float ))
    
  
  }
 }

MPI::Op v468;



void run() {
// Init global cartesian topology
thisRank = MPI::COMM_WORLD.Get_rank();
Nproc = MPI::COMM_WORLD.Get_size();
dimSizes = flocc::findBestDims(Nproc, ndims);

bool periods[ndims]; for (int i = 0; i < ndims; i++) periods[i] = false;
cartComm = MPI::COMM_WORLD.Create_cart(ndims, &dimSizes.front(), periods, true);

{
 int coord[ndims];
for (coord[0] = 0; coord[0] < dimSizes[0]; coord[0]++) {
coordRanks[coord[0]] = cartComm.Get_cart_rank(coord);
}

}
rootRank = coordRanks[0];
v418.Init(&userFun400, true);

v468.Init(&userFun450, true);

  double v1;
  v1 = 0.0;
  double v2;
  v2 = 0.0;
  double v3;
  v3 = 0.0;
  double v4;
  v4 = 0.0;
  int v5;
  v5 = 200000000;
// BEGIN genFloatPairAList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN genFloatPairAList\n";
#endif
#ifdef DEBUG
std::cerr << "WARNING: genFloatPairAList ignores par dims and distributions over ALL nodes.\n"; 
#endif
  struct10 v9;
  
int v6 = v5 / Nproc + (thisRank < (v5%Nproc) ? 1 : 0);
  struct10* v11(new struct10[v6]);
  
for (int v7 = 0; v7 < v6; v7++) {
  double v8 = 2.0; // (double)rand()/RAND_MAX;
    v9.v0 = v8; //(Lf:(Float )) = (Lf:(Float ))
  v9.v1 = v8; //(Lf:(Float )) = (Lf:(Float ))
  
  v11[v7] = v9;
}
// END genFloatPairAList
#ifdef DEBUG
std::cerr << thisRank << ") END genFloatPairAList\n";
#endif
// BEGIN readAList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN readAList\n";
#endif
int v13 = 0;
  // BEGIN reduceList init
    struct10 v14;
      v14.v0 = v4; //(Lf:(Float )) = (Lf:(Float ))
    v14.v1 = v3; //(Lf:(Float )) = (Lf:(Float ))
      struct10 v15(v14);
      struct10 v16;
    
  // END reduceList init
  
if (true) {
  for (int v12 = 0; v12 < v6; v12++) {
      // BEGIN reduceList consume
      double v398;
      v398 = v15.v1 + v11[v12].v1;
      double v399;
      v399 = v15.v0 + v11[v12].v0;
    // TupNd assignment
      v16.v0 = v399; //(Lf:(Float )) = (Lf:(Float ))
      v16.v1 = v398; //(Lf:(Float )) = (Lf:(Float ))
      
      v15 = v16; //((Lf:(Float )), (Lf:(Float ))) = ((Lf:(Float )), (Lf:(Float )))
    
  // END reduceList consume
  
}
}
  // BEGIN reduceList fin
  if (true) {
      v16 = v15; //((Lf:(Float )), (Lf:(Float ))) = ((Lf:(Float )), (Lf:(Float )))
    
    cartComm.Allreduce(&v16, &v15, sizeof(struct10), MPI_PACKED, v418);
  }
  if (cartComm != cartComm) {
    cartComm.Bcast(&v15, sizeof(struct10), MPI_PACKED, rootRank);
  }
  // END reduceList fin
  
// END readAList
#ifdef DEBUG
std::cerr << thisRank << ") END readAList\n";
#endif
// BEGIN countAList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN countAList\n";
#endif
  int v419;
  v419 = v5;
// END countAList
#ifdef DEBUG
std::cerr << thisRank << ") END countAList\n";
#endif
  double v421;
  v421 = (double)v419;
  double v422;
  v422 = v15.v1 / v421;
  double v423;
  v423 = v15.v0 / v421;
// BEGIN readAList
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN readAList\n";
#endif
int v425 = 0;
      // BEGIN reduceList init
        struct10 v430;
          v430.v0 = v2; //(Lf:(Float )) = (Lf:(Float ))
        v430.v1 = v1; //(Lf:(Float )) = (Lf:(Float ))
          struct10 v431(v430);
          struct10 v432;
        
      // END reduceList init
      
if (true) {
  for (int v424 = 0; v424 < v6; v424++) {
      // BEGIN mapList consume:
    double v426;
    double v427;
        double v483;
      v483 = v11[v424].v1 - v422;
      double v484;
      v484 = v11[v424].v0 - v423;
    // TupNd assignment
      v426 = v484; //(Lf:(Float )) = (Lf:(Float ))
      v427 = v483; //(Lf:(Float )) = (Lf:(Float ))
      
    
    // BEGIN mapList consume:
      double v428;
      double v429;
          double v496;
        v496 = v426 * v426;
        double v497;
        v497 = v426 * v427;
      // TupNd assignment
        v428 = v497; //(Lf:(Float )) = (Lf:(Float ))
        v429 = v496; //(Lf:(Float )) = (Lf:(Float ))
        
      
      // BEGIN reduceList consume
          double v448;
          v448 = v431.v1 + v429;
          double v449;
          v449 = v431.v0 + v428;
        // TupNd assignment
          v432.v0 = v449; //(Lf:(Float )) = (Lf:(Float ))
          v432.v1 = v448; //(Lf:(Float )) = (Lf:(Float ))
          
          v431 = v432; //((Lf:(Float )), (Lf:(Float ))) = ((Lf:(Float )), (Lf:(Float )))
        
      // END reduceList consume
      
    // END mapList consume
    
  // END mapList consume
  
}
}
      // BEGIN reduceList fin
      if (true) {
          v432 = v431; //((Lf:(Float )), (Lf:(Float ))) = ((Lf:(Float )), (Lf:(Float )))
        
        cartComm.Allreduce(&v432, &v431, sizeof(struct10), MPI_PACKED, v468);
      }
      if (cartComm != cartComm) {
        cartComm.Bcast(&v431, sizeof(struct10), MPI_PACKED, rootRank);
      }
      // END reduceList fin
      
// END readAList
#ifdef DEBUG
std::cerr << thisRank << ") END readAList\n";
#endif
  double v469;
  v469 = v431.v0 / v431.v1;
  double v470;
  v470 = v423 * v469;
  double v471;
  v471 = v422 - v470;
// BEGIN printFloat
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN printFloat\n";
#endif
if (thisRank == rootRank) std::cout << v471;
// END printFloat
#ifdef DEBUG
std::cerr << thisRank << ") END printFloat\n";
#endif
// BEGIN printFloat
#ifdef DEBUG
std::cerr << thisRank << ") BEGIN printFloat\n";
#endif
if (thisRank == rootRank) std::cout << v469;
// END printFloat
#ifdef DEBUG
std::cerr << thisRank << ") END printFloat\n";
#endif
// No stream called 'final' found in output streams.
}


int main(int argc, char** argv) {
  MPI::Init(argc, argv);
  int rank = MPI::COMM_WORLD.Get_rank();
  int nproc = MPI::COMM_WORLD.Get_size();
  double startTime = MPI::Wtime();

  run();

  double duration = MPI::Wtime() - startTime;
  if (rank == 0) {
    std::cerr << "prog\t" << nproc << "\t" << duration << "\n"; 
  }

  MPI::Finalize();
  return 0;
}
//



