// Manually coded by Tristan Aubrey-Jones 2014
// for comparsion with Flocc generated version
// see http://www.flocc.net/

//#include <vector>
//#include <cstdlib>
//#include <memory>
//#include <cstring> // for strcmp
//#include <cstdio> // sprintf

#include <iostream>
//#include <fstream>
#include <mpi.h>

using namespace std;

typedef int ulong;

#define NV 200000000

// simple linear regression 
// for binary file of 2D double points
int main(int argc, char** argv) {
  // init
  MPI::Init(argc, argv);
  int rank = MPI::COMM_WORLD.Get_rank();
  int N = MPI::COMM_WORLD.Get_size();
  double startTime = MPI::Wtime();

  // make random data
  ulong NLV = (NV / N) + (rank < (NV % N) ? 1 : 0);
  double *data = new double[NLV*2];
  for (ulong i = 0; i < NLV*2; i++) {
    data[i] = 2.0;
  }
  
  // read once, summing data
  double totalX = 0.0, totalY = 0.0;
  for (ulong i = 0; i < NLV*2; i+=2) {
    double x = data[i];
    double y = data[i+1];
    totalX += x; totalY += y;
  }

  // all reduce totals
  double locTotals[2], totals[2];
  locTotals[0] = totalX;
  locTotals[1] = totalY;
  MPI::COMM_WORLD.Allreduce(locTotals, totals, 2, MPI::DOUBLE, MPI::SUM);
  totalX = totals[0];
  totalY = totals[1];

  // calc means
  double avgX = totalX / (double)NV;
  double avgY = totalY / (double)NV;

  // read again, calculating top and bottom of frac
  double top = 0.0, bot = 0.0;
  for (ulong i = 0; i < NLV*2; i+=2) {
    double x = data[i];
    double y = data[i+1];
    top += (x - avgX)*(y - avgY);
    bot += (x - avgX)*(x - avgX);
  }

  // delete data
  delete [] data;

  // gather tops and bots to root
  double sendbuf[2], recvbuf[2];
  sendbuf[0] = top;
  sendbuf[1] = bot;
  MPI::COMM_WORLD.Reduce(sendbuf, recvbuf, 2, MPI::DOUBLE,
    MPI::SUM, 0);

  if (rank == 0) {
    // unpack result
    top = recvbuf[0];
    bot = recvbuf[1];

    // calc coeffs
    double beta = (bot == 0.0) ? 1234.0 : (top / bot);
    double alpha = avgY - (avgX * beta);

    // display
    double endTime = MPI::Wtime();
    double dur = endTime - startTime;
    cout << "avg x = " << avgX << ", avg y = " << avgY << "\n";
    cout << "y = " << alpha << " + " << beta << "x\n";
    cerr << "linefitm\t" << NV << "\t" << N << "\t" << dur << "\n"; 
  }

  // end
  MPI::Finalize();
  return 0;
}

